/* Special Initializers for certain USB Mass Storage devices
 *
 * $Id: initializers.c,v 1.2 2000/09/06 22:35:57 mdharm Exp $
 *
 * Current development and maintenance by:
 *   (c) 1999, 2000 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 *
 * This driver is based on the 'USB Mass Storage Class' document. This
 * describes in detail the protocol used to communicate with such
 * devices.  Clearly, the designers had SCSI and ATAPI commands in
 * mind when they created this document.  The commands are all very
 * similar to commands in the SCSI-II and ATAPI specifications.
 *
 * It is important to note that in a number of cases this class
 * exhibits class-specific exemptions from the USB specification.
 * Notably the usage of NAK, STALL and ACK differs from the norm, in
 * that they are used to communicate wait, failed and OK on commands.
 *
 * Also, for certain devices, the interrupt endpoint is used to convey
 * status of a command.
 *
 * Please see http://www.one-eyed-alien.net/~mdharm/linux-usb for more
 * information about this driver.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "initializers.h"
#include "debug.h"
#include "transport.h"

/* This places the Shuttle/SCM USB<->SCSI bridge devices in multi-target
 * mode */
int usb_stor_euscsi_init(struct us_data *us)
{
	unsigned char data = 0x1;
	int result;

	US_DEBUGP("Attempting to init eUSCSI bridge...\n");
	result = usb_control_msg(us->pusb_dev, usb_sndctrlpipe(us->pusb_dev, 0),
			0x0C, USB_RECIP_INTERFACE | USB_TYPE_VENDOR,
			0x01, 0x0, &data, 0x1, 5*HZ);
	US_DEBUGP("-- result is %d\n", result);
	US_DEBUGP("-- data afterwards is %d\n", data);

	return 0;
}

/* This function is required to activate all four slots on the UCR-61S2B
 * flash reader */

int usb_stor_ucr61s2b_init(struct us_data *us)
{
	int pipe;
	struct bulk_cb_wrap *bcb;
	struct bulk_cs_wrap *bcs;
	int res, partial;

	bcb = kmalloc(sizeof *bcb, in_interrupt() ? GFP_ATOMIC : GFP_NOIO);
	if (!bcb) {
		return(-1);
	}
	bcs = kmalloc(sizeof *bcs, in_interrupt() ? GFP_ATOMIC : GFP_NOIO);
	if (!bcs) {
		kfree(bcb);
		return(-1);
	}

	US_DEBUGP("Sending UCR-61S2B initialization packet...\n");

	bcb->Signature = cpu_to_le32(US_BULK_CB_SIGN);
	bcb->Tag = ++(us->tag);
	bcb->DataTransferLength = cpu_to_le32(0);
	bcb->Flags = bcb->Lun = 0;
	bcb->Length = sizeof(UCR61S2B_INIT);
	memset(bcb->CDB, 0, sizeof(bcb->CDB));
	memcpy(bcb->CDB, UCR61S2B_INIT, sizeof(UCR61S2B_INIT));

	pipe = usb_sndbulkpipe(us->pusb_dev, us->ep_out);
	res = usb_stor_bulk_msg(us, bcb, pipe, US_BULK_CB_WRAP_LEN, &partial);
	US_DEBUGP("-- result is %d\n", res);
	kfree(bcb);

	if(res) {
		kfree(bcs);
		return(res);
	}

	pipe = usb_rcvbulkpipe(us->pusb_dev, us->ep_in);
	res = usb_stor_bulk_msg(us, bcs, pipe, US_BULK_CS_WRAP_LEN, &partial);
	US_DEBUGP("-- result of status read is %d\n", res);

	kfree(bcs);

	return(res ? -1 : 0);
}

/* This places the HUAWEI E220 devices in multi-port mode */
int usb_stor_switch_ports_init(struct us_data *us)
{
       int result = 0;
       unsigned char data = 0x1;
       //us->iobuf[0] = 0x1;
       result = usb_control_msg(us->pusb_dev, usb_sndctrlpipe(us->pusb_dev, 0),
               USB_REQ_SET_FEATURE, USB_TYPE_STANDARD | USB_RECIP_DEVICE,
               0x01, 0x0, &data, 0x1, 1000);
               //0x01, 0x0, us->iobuf, 0x1, 1000);
       US_DEBUGP("usb_control_msg performing result is %d\n", result);
       return (result ? 0 : -1);
}
#ifdef HSDPA
#define ZCD_FORCE_MODEM                 0x01
#define ZCD_ALLOW_MS                    0x02
#define RESPONSE_LEN 1024

static unsigned int option_zero_cd = ZCD_FORCE_MODEM;

static int option_rezero(struct us_data *us, int ep_in, int ep_out)
{
        const unsigned char rezero_msg[] = {
          0x55, 0x53, 0x42, 0x43, 0x78, 0x56, 0x34, 0x12,
          0x01, 0x00, 0x00, 0x00, 0x80, 0x00, 0x06, 0x01,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        };
        char *buffer;
        int result;

        US_DEBUGP("Option MS: %s", "DEVICE MODE SWITCH\n",__FUNCTION__);

        //buffer = kzalloc(RESPONSE_LEN, GFP_KERNEL);
        buffer = (char *)kmalloc(sizeof(rezero_msg), GFP_KERNEL);

        if (buffer == NULL)
                return USB_STOR_TRANSPORT_ERROR;

        memcpy(buffer, rezero_msg, sizeof (rezero_msg));
        result = usb_stor_bulk_transfer_buf(us,
                        usb_sndbulkpipe(us->pusb_dev, ep_out),
                        buffer, sizeof (rezero_msg), NULL);
        if (result != USB_STOR_XFER_GOOD) {
                result = USB_STOR_XFER_ERROR;
                goto out;
        }
        /* Some of the devices need to be asked for a response, but we don't
         * care what that response is.
         */
        result = usb_stor_bulk_transfer_buf(us,
                        usb_sndbulkpipe(us->pusb_dev, ep_out),
                        buffer, RESPONSE_LEN, NULL);
        result = USB_STOR_XFER_GOOD;

out:
        kfree(buffer);
        return result;
}
static int option_ZTE626_rezero(struct us_data *us, int ep_in, int ep_out)
{
#if 0
        const unsigned char rezero_msg[] = {
          0x55,0x53,0x42,0x43,0x12,0x34,0x56,0x78,0x00,0x00,0x00,
    0x00,0x00,0x00,0x06,0x1b,0x00,0x00,0x00,0x03,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
        };
#endif
        const unsigned char rezero_msg[] = {
        0x55,0x53,0x42,0x43,0x12,0x34,0x56,0x78,
  0x20,0x00,0x00,0x00,0x80,0x00,0x0c,0x85,
  0x01,0x01,0x01,0x18,0x01,0x01,0x01,0x01,
  0x01,0x00,0x00,0x00,0x00,0x00,0x00
  };

        char *buffer;
        int result;

        US_DEBUGP("Option MS: %s", "DEVICE MODE SWITCH\n",__FUNCTION__);

        //buffer = kzalloc(RESPONSE_LEN, GFP_KERNEL);
        buffer = (char *)kmalloc(sizeof(rezero_msg), GFP_KERNEL);

        if (buffer == NULL)
                return USB_STOR_TRANSPORT_ERROR;

        memcpy(buffer, rezero_msg, sizeof (rezero_msg));
        result = usb_stor_bulk_transfer_buf(us,
                        usb_sndbulkpipe(us->pusb_dev, ep_out),
                        buffer, sizeof (rezero_msg), NULL);
        printk("%s(%d)\n",__FUNCTION__,__LINE__);
        if (result != USB_STOR_XFER_GOOD) {
                result = USB_STOR_XFER_ERROR;
                US_DEBUGP("Option MS: line 430 %s", "USB_STOR_XFER_ERROR\n",__FUNCTION__);
                goto out;
        }

        /* Some of the devices need to be asked for a response, but we don't
         * care what that response is.
         */
        printk("%s(%d)\n",__FUNCTION__,__LINE__);
        #if 0
        result = usb_stor_bulk_transfer_buf(us,
                        usb_sndbulkpipe(us->pusb_dev, ep_out),
                        buffer, RESPONSE_LEN, NULL);
        #endif
        printk("%s(%d)\n",__FUNCTION__,__LINE__);
        result = USB_STOR_XFER_GOOD;

out:
        kfree(buffer);
        return result;
}

int usb_stor_zte_mf626_init(struct us_data *us)
 {
  struct usb_device *udev;
  struct usb_interface *intf;
  struct usb_host_interface *iface_desc;
  struct usb_endpoint_descriptor *endpoint = NULL;
  u8 ep_in = 0, ep_out = 0;
  int ep_in_size = 0, ep_out_size = 0;
  int i, result;

#if 0
  udev = us->pusb_dev;
  intf = us->pusb_intf;

  /* Ensure it's really a ZeroCD device; devices that are already
   * in modem mode return 0xFF for class, subclass, and protocol.
   */
  if (udev->descriptor.bDeviceClass != 0 ||
      udev->descriptor.bDeviceSubClass != 0 ||
      udev->descriptor.bDeviceProtocol != 0)
   return USB_STOR_TRANSPORT_GOOD;

  US_DEBUGP("ZTE MS: option_ms_init called\n");

  /* Find the right mass storage interface */
  iface_desc = intf->cur_altsetting;
  if (iface_desc->desc.bInterfaceClass != 0x8 ||
      iface_desc->desc.bInterfaceSubClass != 0x5 ||
      iface_desc->desc.bInterfaceProtocol != 0x50) {
   US_DEBUGP("ZTE MS: mass storage interface not found, no action "
             "required\n");
   return USB_STOR_TRANSPORT_GOOD;
  }

  /* Find the mass storage bulk endpoints */
  for (i = 0; i < iface_desc->desc.bNumEndpoints && (!ep_in_size || !ep_out_size); ++i) {
   endpoint = &iface_desc->endpoint[i].desc;

   if (usb_endpoint_is_bulk_in(endpoint)) {
    ep_in = usb_endpoint_num(endpoint);
    //ep_in_size = le16_to_cpu(endpoint->wMaxPacketSize);
    ep_in_size = (endpoint->wMaxPacketSize > maxSize)?endpoint->wMaxPacketSize:maxSize;
   } else if (usb_endpoint_is_bulk_out(endpoint)) {
    ep_out = usb_endpoint_num(endpoint);
    //ep_out_size = le16_to_cpu(endpoint->wMaxPacketSize);
    ep_out_size = (endpoint->wMaxPacketSize > maxSize)?endpoint->wMaxPacketSize:maxSize;
   }
  }

  /* Can't find the mass storage endpoints */
  if (!ep_in_size || !ep_out_size) {
   US_DEBUGP("ZTE MS: mass storage endpoints not found, no action "
             "required\n");
   return USB_STOR_TRANSPORT_GOOD;
  }
#endif
  /* Force Modem mode */
  if (option_zero_cd == ZCD_FORCE_MODEM) {
   US_DEBUGP("ZTE MS: %s", "Forcing Modem Mode\n",__FUNCTION__);
   result = option_ZTE626_rezero(us, us->ep_in, us->ep_out);
   if (result != USB_STOR_XFER_GOOD)
    US_DEBUGP("ZTE MS: Failed to switch to modem mode.\n");
   return -EIO;
  } else if (option_zero_cd == ZCD_ALLOW_MS) {
   /* Allow Mass Storage mode (keep CD-Rom) */
   US_DEBUGP("ZTE MS: %s", "Allowing Mass Storage Mode if device"
             " requests it\n",__FUNCTION__);
  }

  return USB_STOR_TRANSPORT_GOOD;
 }
#endif

